/* sqUnixMemory -- allocation routines for Squeak's object memory
 *
 */


#include <sq.h>


/* whether to allocate by mmap() on /dev/zero or by malloc().  The
   former is less portable, but it allows the allocated region to grow
   in place */

#define MEMORY_USE_MMAP   /* this should be moved into a config file */

#ifdef  MEMORY_USE_MMAP

#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>


/* max memory to allocate.  again, this should be configurable... */
#define MEMORY_TO_USE (1024*1024*500)   /* 500 megs */


static void *heap = NULL;  /* location of the heap */
static int amountInUse = 0;  /* amount currently in use */


/* return the last byte of accessible heap */
static void *heapLimit() 
{
  return (char *)heap + amountInUse - 1;
}


void * sqAllocateMemory(int minHeapSize, int desiredHeapSize)  {
  int heapFd;

  if(heap != NULL) {
    fprintf(stderr, "sqAllocateMemory called twice!\n");
    exit(1);
  }

  if(desiredHeapSize > MEMORY_TO_USE) {
    fprintf(stderr, "requested %d memory for heap!  Giving up.\n", desiredHeapSize);
    return NULL;
  }
  
  heapFd = open("/dev/zero", O_RDWR);   /* XXX add includes for this */
  if(heapFd < 0)
    return NULL;
  
  
  heap = mmap(NULL, MEMORY_TO_USE, PROT_READ|PROT_WRITE, MAP_PRIVATE, heapFd, 0);
  
  if(heap == NULL) {
    perror("sqAllocateMemory: mmap");
    close(heapFd);
    return NULL;
  }

  close(heapFd);
  amountInUse = desiredHeapSize;
  return heap;
}


void * sqGrowMemoryBy(void *oldLimit, int delta) {
#if 0
  printf("growing by %d...\n", delta);
#endif
  
  amountInUse += delta;
  if(amountInUse > MEMORY_TO_USE) {
    /* requested too much */
    amountInUse = MEMORY_TO_USE;
  }
  
  return heapLimit();
}


void * sqShrinkMemoryBy(void *oldLimit, int delta) {
#if 0
  printf("shrinking by %d...\n", delta);
#endif
  
  amountInUse -= delta;
  if(amountInUse < 0) {
    fprintf(stderr, "heap shrunk to less than 0 bytes!\n");
    exit(1);
  }
  
  return heapLimit();
}

int sqMemoryExtraBytesLeft(int includingSwap) {
  return MEMORY_TO_USE - amountInUse;
}


#else


void * sqAllocateMemory(int minHeapSize, int desiredHeapSize) {
  malloc(desiredHeapSize);
}

void * sqGrowMemoryBy(void *oldLimit, int delta) {
  return oldLimit;
}

void * sqShrinkMemoryBy(void *oldLimit, int delta) {
  return oldLimit;
}

int sqMemoryExtraBytesLeft(int includingSwap) {
  return 0;
}


#endif
